@@ -1,11 +1,9 @@
-# coding: utf-8
 # frozen_string_literal: true
 
 require 'rubygems/package/tar_test_case'
-require 'rubygems/simple_gem'
+require 'digest'
 
 class TestGemPackage < Gem::Package::TarTestCase
-
   def setup
     super
 
@@ -24,6 +22,8 @@ class TestGemPackage < Gem::Package::TarTestCase
   end
 
   def test_class_new_old_format
+    skip "jruby can't require the simple_gem file" if Gem.java_platform?
+    require_relative "simple_gem"
     File.open 'old_format.gem', 'wb' do |io|
       io.write SIMPLE_GEM
     end
@@ -91,27 +91,68 @@ class TestGemPackage < Gem::Package::TarTestCase
       'SHA512' => {
         'metadata.gz' => metadata_sha512,
         'data.tar.gz' => Digest::SHA512.hexdigest(tar),
-      }
-    }
-
-    if defined?(OpenSSL::Digest) then
-      expected['SHA256'] = {
+      },
+      'SHA256' => {
         'metadata.gz' => metadata_sha256,
         'data.tar.gz' => Digest::SHA256.hexdigest(tar),
-      }
-    end
+      },
+    }
 
     assert_equal expected, YAML.load(checksums)
   end
 
+  def test_build_time_uses_source_date_epoch
+    epoch = ENV["SOURCE_DATE_EPOCH"]
+    ENV["SOURCE_DATE_EPOCH"] = "123456789"
+
+    spec = Gem::Specification.new 'build', '1'
+    spec.summary = 'build'
+    spec.authors = 'build'
+    spec.files = ['lib/code.rb']
+    spec.date = Time.at 0
+    spec.rubygems_version = Gem::Version.new '0'
+
+    package = Gem::Package.new spec.file_name
+
+    assert_equal Time.at(ENV["SOURCE_DATE_EPOCH"].to_i).utc, package.build_time
+  ensure
+    ENV["SOURCE_DATE_EPOCH"] = epoch
+  end
+
+  def test_build_time_without_source_date_epoch
+    epoch = ENV["SOURCE_DATE_EPOCH"]
+    ENV["SOURCE_DATE_EPOCH"] = nil
+
+    spec = Gem::Specification.new 'build', '1'
+    spec.summary = 'build'
+    spec.authors = 'build'
+    spec.files = ['lib/code.rb']
+    spec.rubygems_version = Gem::Version.new '0'
+
+    package = Gem::Package.new spec.file_name
+
+    assert_kind_of Time, package.build_time
+
+    build_time = package.build_time.to_i
+
+    assert_equal Gem.source_date_epoch.to_i, build_time
+  ensure
+    ENV["SOURCE_DATE_EPOCH"] = epoch
+  end
+
   def test_add_files
     spec = Gem::Specification.new
     spec.files = %w[lib/code.rb lib/empty]
 
     FileUtils.mkdir_p 'lib/empty'
 
-    File.open 'lib/code.rb',  'w' do |io| io.write '# lib/code.rb'  end
-    File.open 'lib/extra.rb', 'w' do |io| io.write '# lib/extra.rb' end
+    File.open 'lib/code.rb', 'w' do |io|
+      io.write '# lib/code.rb'
+    end
+
+    File.open 'lib/extra.rb', 'w' do |io|
+      io.write '# lib/extra.rb'
+    end
 
     package = Gem::Package.new 'bogus.gem'
     package.spec = spec
@@ -134,16 +175,26 @@ class TestGemPackage < Gem::Package::TarTestCase
   end
 
   def test_add_files_symlink
-    skip 'symlink not supported' if Gem.win_platform?
-
     spec = Gem::Specification.new
-    spec.files = %w[lib/code.rb lib/code_sym.rb]
+    spec.files = %w[lib/code.rb lib/code_sym.rb lib/code_sym2.rb]
 
     FileUtils.mkdir_p 'lib'
-    File.open 'lib/code.rb',  'w' do |io| io.write '# lib/code.rb'  end
+
+    File.open 'lib/code.rb', 'w' do |io|
+      io.write '# lib/code.rb'
+    end
 
     # NOTE: 'code.rb' is correct, because it's relative to lib/code_sym.rb
-    File.symlink('code.rb', 'lib/code_sym.rb')
+    begin
+      File.symlink('code.rb', 'lib/code_sym.rb')
+      File.symlink('../lib/code.rb', 'lib/code_sym2.rb')
+    rescue Errno::EACCES => e
+      if win_platform?
+        skip "symlink - must be admin with no UAC on Windows"
+      else
+        raise e
+      end
+    end
 
     package = Gem::Package.new 'bogus.gem'
     package.spec = spec
@@ -167,7 +218,7 @@ class TestGemPackage < Gem::Package::TarTestCase
     end
 
     assert_equal %w[lib/code.rb], files
-    assert_equal [{'lib/code_sym.rb' => 'lib/code.rb'}], symlinks
+    assert_equal [{'lib/code_sym.rb' => 'code.rb'}, {'lib/code_sym2.rb' => '../lib/code.rb'}], symlinks
   end
 
   def test_build
@@ -201,7 +252,7 @@ class TestGemPackage < Gem::Package::TarTestCase
   end
 
   def test_build_auto_signed
-    skip 'openssl is missing' unless defined?(OpenSSL::SSL)
+    skip 'openssl is missing' unless Gem::HAVE_OPENSSL
 
     FileUtils.mkdir_p File.join(Gem.user_home, '.gem')
 
@@ -244,7 +295,7 @@ class TestGemPackage < Gem::Package::TarTestCase
   end
 
   def test_build_auto_signed_encrypted_key
-    skip 'openssl is missing' unless defined?(OpenSSL::SSL)
+    skip 'openssl is missing' unless Gem::HAVE_OPENSSL
 
     FileUtils.mkdir_p File.join(Gem.user_home, '.gem')
 
@@ -299,8 +350,21 @@ class TestGemPackage < Gem::Package::TarTestCase
     assert_equal 'missing value for attribute summary', e.message
   end
 
+  def test_build_invalid_arguments
+    spec = Gem::Specification.new 'build', '1'
+
+    package = Gem::Package.new spec.file_name
+    package.spec = spec
+
+    e = assert_raises ArgumentError do
+      package.build true, true
+    end
+
+    assert_equal "skip_validation = true and strict_validation = true are incompatible", e.message
+  end
+
   def test_build_signed
-    skip 'openssl is missing' unless defined?(OpenSSL::SSL)
+    skip 'openssl is missing' unless Gem::HAVE_OPENSSL
 
     spec = Gem::Specification.new 'build', '1'
     spec.summary = 'build'
@@ -337,7 +401,7 @@ class TestGemPackage < Gem::Package::TarTestCase
   end
 
   def test_build_signed_encrypted_key
-    skip 'openssl is missing' unless defined?(OpenSSL::SSL)
+    skip 'openssl is missing' unless Gem::HAVE_OPENSSL
 
     spec = Gem::Specification.new 'build', '1'
     spec.summary = 'build'
@@ -373,6 +437,33 @@ class TestGemPackage < Gem::Package::TarTestCase
     assert_equal %w[lib/code.rb], reader.contents
   end
 
+  def test_raw_spec
+    data_tgz = util_tar_gz {}
+
+    gem = util_tar do |tar|
+      tar.add_file 'data.tar.gz', 0644 do |io|
+        io.write data_tgz.string
+      end
+
+      tar.add_file 'metadata.gz', 0644 do |io|
+        Zlib::GzipWriter.wrap io do |gzio|
+          gzio.write @spec.to_yaml
+        end
+      end
+    end
+
+    gem_path = "#{@destination}/test.gem"
+
+    File.open gem_path, "wb" do |io|
+      io.write gem.string
+    end
+
+    spec, metadata = Gem::Package.raw_spec(gem_path)
+
+    assert_equal @spec, spec
+    assert_match @spec.to_yaml, metadata.force_encoding("UTF-8")
+  end
+
   def test_contents
     package = Gem::Package.new @gem
 
@@ -394,7 +485,7 @@ class TestGemPackage < Gem::Package::TarTestCase
   end
 
   def test_extract_files_empty
-    data_tgz = util_tar_gz do end
+    data_tgz = util_tar_gz {}
 
     gem = util_tar do |tar|
       tar.add_file 'data.tar.gz', 0644 do |io|
@@ -423,7 +514,9 @@ class TestGemPackage < Gem::Package::TarTestCase
     package = Gem::Package.new @gem
 
     tgz_io = util_tar_gz do |tar|
-      tar.add_file '/absolute.rb', 0644 do |io| io.write 'hi' end
+      tar.add_file '/absolute.rb', 0644 do |io|
+        io.write 'hi'
+      end
     end
 
     e = assert_raises Gem::Package::PathError do
@@ -435,17 +528,26 @@ class TestGemPackage < Gem::Package::TarTestCase
   end
 
   def test_extract_tar_gz_symlink_relative_path
-    skip 'symlink not supported' if Gem.win_platform?
-
     package = Gem::Package.new @gem
 
     tgz_io = util_tar_gz do |tar|
-      tar.add_file    'relative.rb', 0644 do |io| io.write 'hi' end
-      tar.mkdir       'lib',         0755
+      tar.add_file 'relative.rb', 0644 do |io|
+        io.write 'hi'
+      end
+
+      tar.mkdir       'lib', 0755
       tar.add_symlink 'lib/foo.rb', '../relative.rb', 0644
     end
 
-    package.extract_tar_gz tgz_io, @destination
+    begin
+      package.extract_tar_gz tgz_io, @destination
+    rescue Errno::EACCES => e
+      if win_platform?
+        skip "symlink - must be admin with no UAC on Windows"
+      else
+        raise e
+      end
+    end
 
     extracted = File.join @destination, 'lib/foo.rb'
     assert_path_exists extracted
@@ -456,33 +558,37 @@ class TestGemPackage < Gem::Package::TarTestCase
   end
 
   def test_extract_symlink_parent
-   skip 'symlink not supported' if Gem.win_platform?
+    package = Gem::Package.new @gem
 
-   package = Gem::Package.new @gem
+    tgz_io = util_tar_gz do |tar|
+      tar.mkdir       'lib',               0755
+      tar.add_symlink 'lib/link', '../..', 0644
+      tar.add_file    'lib/link/outside.txt', 0644 do |io|
+        io.write 'hi'
+      end
+    end
 
-   tgz_io = util_tar_gz do |tar|
-     tar.mkdir       'lib',               0755
-     tar.add_symlink 'lib/link', '../..', 0644
-     tar.add_file    'lib/link/outside.txt', 0644 do |io| io.write 'hi' end
-   end
+    # Extract into a subdirectory of @destination; if this test fails it writes
+    # a file outside destination_subdir, but we want the file to remain inside
+    # @destination so it will be cleaned up.
+    destination_subdir = File.join @destination, 'subdir'
+    FileUtils.mkdir_p destination_subdir
 
-   # Extract into a subdirectory of @destination; if this test fails it writes
-   # a file outside destination_subdir, but we want the file to remain inside
-   # @destination so it will be cleaned up.
-   destination_subdir = File.join @destination, 'subdir'
-   FileUtils.mkdir_p destination_subdir
+    e = assert_raises(Gem::Package::PathError, Errno::EACCES) do
+      package.extract_tar_gz tgz_io, destination_subdir
+    end
 
-   e = assert_raises Gem::Package::PathError do
-     package.extract_tar_gz tgz_io, destination_subdir
-   end
-
-   assert_equal("installing into parent path lib/link/outside.txt of " +
-                 "#{destination_subdir} is not allowed", e.message)
+    if Gem::Package::PathError === e
+      assert_equal("installing into parent path lib/link/outside.txt of " +
+                  "#{destination_subdir} is not allowed", e.message)
+    elsif win_platform?
+      skip "symlink - must be admin with no UAC on Windows"
+    else
+      raise e
+    end
   end
 
   def test_extract_symlink_parent_doesnt_delete_user_dir
-    skip if RUBY_VERSION <= "1.8.7"
-
     package = Gem::Package.new @gem
 
     # Extract into a subdirectory of @destination; if this test fails it writes
@@ -495,6 +601,8 @@ class TestGemPackage < Gem::Package::TarTestCase
     destination_user_subdir = File.join destination_user_dir, 'dir'
     FileUtils.mkdir_p destination_user_subdir
 
+    skip "TMPDIR seems too long to add it as symlink into tar" if destination_user_dir.size > 90
+
     tgz_io = util_tar_gz do |tar|
       tar.add_symlink 'link', destination_user_dir, 16877
       tar.add_symlink 'link/dir', '.', 16877
@@ -521,8 +629,10 @@ class TestGemPackage < Gem::Package::TarTestCase
 
     tgz_io = util_tar_gz do |tar|
       tar.mkdir    'lib',        0755
-      tar.add_file 'lib/foo.rb', 0644 do |io| io.write 'hi' end
-      tar.mkdir    'lib/foo',    0755
+      tar.add_file 'lib/foo.rb', 0644 do |io|
+        io.write 'hi'
+      end
+      tar.mkdir    'lib/foo', 0755
     end
 
     package.extract_tar_gz tgz_io, @destination
@@ -538,7 +648,9 @@ class TestGemPackage < Gem::Package::TarTestCase
     package = Gem::Package.new @gem
 
     tgz_io = util_tar_gz do |tar|
-      tar.add_file './dot_slash.rb', 0644 do |io| io.write 'hi' end
+      tar.add_file './dot_slash.rb', 0644 do |io|
+        io.write 'hi'
+      end
     end
 
     package.extract_tar_gz tgz_io, @destination
@@ -551,7 +663,9 @@ class TestGemPackage < Gem::Package::TarTestCase
     package = Gem::Package.new @gem
 
     tgz_io = util_tar_gz do |tar|
-      tar.add_file '.dot_file.rb', 0644 do |io| io.write 'hi' end
+      tar.add_file '.dot_file.rb', 0644 do |io|
+        io.write 'hi'
+      end
     end
 
     package.extract_tar_gz tgz_io, @destination
@@ -565,7 +679,9 @@ class TestGemPackage < Gem::Package::TarTestCase
       package = Gem::Package.new @gem
 
       tgz_io = util_tar_gz do |tar|
-        tar.add_file 'foo/file.rb', 0644 do |io| io.write 'hi' end
+        tar.add_file 'foo/file.rb', 0644 do |io|
+          io.write 'hi'
+        end
       end
 
       package.extract_tar_gz tgz_io, @destination.upcase
@@ -579,12 +695,12 @@ class TestGemPackage < Gem::Package::TarTestCase
     package = Gem::Package.new @gem
 
     file = 'file.rb'.dup
-    file.taint
+    file.taint if RUBY_VERSION < '2.7'
 
     destination = package.install_location file, @destination
 
     assert_equal File.join(@destination, 'file.rb'), destination
-    refute destination.tainted?
+    refute destination.tainted? if RUBY_VERSION < '2.7'
   end
 
   def test_install_location_absolute
@@ -615,18 +731,17 @@ class TestGemPackage < Gem::Package::TarTestCase
   end
 
   def test_install_location_extra_slash
-    skip 'no File.realpath on 1.8' if RUBY_VERSION < '1.9'
     package = Gem::Package.new @gem
 
     file = 'foo//file.rb'.dup
-    file.taint
+    file.taint if RUBY_VERSION < '2.7'
 
     destination = @destination.sub '/', '//'
 
     destination = package.install_location file, destination
 
     assert_equal File.join(@destination, 'foo', 'file.rb'), destination
-    refute destination.tainted?
+    refute destination.tainted? if RUBY_VERSION < '2.7'
   end
 
   def test_install_location_relative
@@ -771,6 +886,7 @@ class TestGemPackage < Gem::Package::TarTestCase
   end
 
   def test_verify_corrupt
+    skip "jruby strips the null byte and does not think it's corrupt" if Gem.java_platform?
     tf = Tempfile.open 'corrupt' do |io|
       data = Gem::Util.gzip 'a' * 10
       io.write \
@@ -788,7 +904,7 @@ class TestGemPackage < Gem::Package::TarTestCase
                    e.message
       io
     end
-    tf.close! if tf.respond_to? :close!
+    tf.close!
   end
 
   def test_verify_empty
@@ -810,8 +926,8 @@ class TestGemPackage < Gem::Package::TarTestCase
       package.verify
     end
 
-    assert_match %r%^No such file or directory%, e.message
-    assert_match %r%nonexistent.gem$%,           e.message
+    assert_match %r{^No such file or directory}, e.message
+    assert_match %r{nonexistent.gem$},           e.message
   end
 
   def test_verify_duplicate_file
@@ -841,7 +957,7 @@ class TestGemPackage < Gem::Package::TarTestCase
   end
 
   def test_verify_security_policy
-    skip 'openssl is missing' unless defined?(OpenSSL::SSL)
+    skip 'openssl is missing' unless Gem::HAVE_OPENSSL
 
     package = Gem::Package.new @gem
     package.security_policy = Gem::Security::HighSecurity
@@ -858,7 +974,7 @@ class TestGemPackage < Gem::Package::TarTestCase
   end
 
   def test_verify_security_policy_low_security
-    skip 'openssl is missing' unless defined?(OpenSSL::SSL)
+    skip 'openssl is missing' unless Gem::HAVE_OPENSSL
 
     @spec.cert_chain = [PUBLIC_CERT.to_pem]
     @spec.signing_key = PRIVATE_KEY
@@ -878,7 +994,7 @@ class TestGemPackage < Gem::Package::TarTestCase
   end
 
   def test_verify_security_policy_checksum_missing
-    skip 'openssl is missing' unless defined?(OpenSSL::SSL)
+    skip 'openssl is missing' unless Gem::HAVE_OPENSSL
 
     @spec.cert_chain = [PUBLIC_CERT.to_pem]
     @spec.signing_key = PRIVATE_KEY
@@ -899,7 +1015,7 @@ class TestGemPackage < Gem::Package::TarTestCase
         bogus_data = Gem::Util.gzip 'hello'
         fake_signer = Class.new do
           def digest_name; 'SHA512'; end
-          def digest_algorithm; Digest(:SHA512); end
+          def digest_algorithm; Digest(:SHA512).new; end
           def key; 'key'; end
           def sign(*); 'fake_sig'; end
         end
@@ -949,11 +1065,22 @@ class TestGemPackage < Gem::Package::TarTestCase
 
     package = Gem::Package.new @gem
 
-    e = assert_raises Gem::Package::FormatError do
-      package.verify_entry entry
+    _, err = use_ui @ui do
+      e = nil
+
+      out_err = capture_io do
+        e = assert_raises ArgumentError do
+          package.verify_entry entry
+        end
+      end
+
+      assert_equal "whatever", e.message
+      assert_equal "full_name", e.backtrace_locations.first.label
+
+      out_err
     end
 
-    assert_equal "package is corrupt, exception while verifying: whatever (ArgumentError) in #{@gem}", e.message
+    assert_equal "Exception while verifying #{@gem}\n", err
 
     valid_metadata = ["metadata", "metadata.gz"]
     valid_metadata.each do |vm|
@@ -978,7 +1105,7 @@ class TestGemPackage < Gem::Package::TarTestCase
       $bad_name = vm
 
       entry = Object.new
-      def entry.full_name() $bad_name  end
+      def entry.full_name() $bad_name end
 
       package = Gem::Package.new(@gem)
       package.instance_variable_set(:@files, [])
@@ -996,6 +1123,11 @@ class TestGemPackage < Gem::Package::TarTestCase
     assert_equal @spec, package.spec
   end
 
+  def test_gem_attr
+    package = Gem::Package.new(@gem)
+    assert_equal(@gem, package.gem.path)
+  end
+
   def test_spec_from_io
     # This functionality is used by rubygems.org to extract spec data from an
     # uploaded gem before it is written to storage.
@@ -1031,9 +1163,10 @@ class TestGemPackage < Gem::Package::TarTestCase
     tgz_io = StringIO.new
 
     # can't wrap TarWriter because it seeks
-    Zlib::GzipWriter.wrap tgz_io do |io| io.write tar_io.string end
+    Zlib::GzipWriter.wrap tgz_io do |io|
+      io.write tar_io.string
+    end
 
     StringIO.new tgz_io.string
   end
-
 end
